/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2013
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev$
 * $Date$
 * $Author$
 * 
 */

package de.unihalle.informatik.Alida.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.*;

import net.java.sezpoz.Index;
import net.java.sezpoz.IndexItem;

import de.unihalle.informatik.Alida.operator.*;
import de.unihalle.informatik.Alida.operator.events.ALDConfigurationEvent;
import de.unihalle.informatik.Alida.workflows.ALDWorkflow;
import de.unihalle.informatik.Alida.workflows.ALDWorkflowNode.ALDWorkflowNodeState;
import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.indexing.SezPozAdapter;
import de.unihalle.informatik.Alida.batch.ALDBatchInputManagerSwing;
import de.unihalle.informatik.Alida.batch.ALDBatchOutputManagerSwing;
import de.unihalle.informatik.Alida.batch.provider.input.swing.ALDBatchInputIteratorSwing;
import de.unihalle.informatik.Alida.dataio.provider.swing.components.ALDSwingComponent;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.*;
import de.unihalle.informatik.Alida.exceptions.ALDBatchIOException;
import de.unihalle.informatik.Alida.exceptions.ALDBatchIOManagerException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDWorkflowException;
import de.unihalle.informatik.Alida.gui.ALDOperatorGUIExecutionProxy.WorkflowThreadStatus;

/**
 * Frame to configure and run an operator.
 * <p>
 * If the operator is properly configured, i.e. its input parameters are set, it
 * is executed as an {@link ALDWorkflow} containing only a single node. 
 * Result display and interaction with the operator are handled by the event 
 * mechanisms of the workflow. 
 * <p>
 * Controllable operators trigger the display of additional buttons in GUI for
 * pausing, resuming and interrupting calculations.
 * 
 * @author Stefan Posch
 * @author Birgit Moeller
 */
public class ALDOperatorControlFrame extends ALDOperatorConfigurationFrame 
	implements ItemListener {

	/**
	 * Local flag for debug output.
	 */
	@SuppressWarnings("unused")
	private boolean debug = false;

	
	/* ***********************************
	 * Some local data type declarations.
	 * ***********************************/
  
	/**
	 * Default run button color of unconfigured operators. 
	 */
	protected final static Color buttonColor_unconfigured = 
			new Color(255, 94, 94);
	
	/**
	 * Default run button color of runnable operators. 
	 */
	protected final static Color buttonColor_runnable =
			new Color(255, 255, 122);
	
	/**
	 * Default run button color of running operators. 
	 */
	protected final static Color buttonColor_running =
			new Color(106, 106, 214);

	/**
	 * Default run button color of ready operators. 
	 */
	protected final static Color buttonColor_ready =
			new Color(118, 200, 118);

	/**
	 * Reference to the surrounding workflow manager.
	 */
	protected ALDOperatorGUIExecutionProxy execProxy = null;
	
	/* **********************************************************************
	 * Buttons and other components for interactions with running operators.
	 * **********************************************************************/

	/**
	 * Result display button.
	 */
	protected JButton resultDisplayButton;

	/**
	 * Run button.
	 */
	protected JButton runButton;
	
	/**
	 * Pause button, changes its label.
	 */
	protected JButton pauseButton;

	/**
	 * Stop button to stop snake optimization (after end of iteration).
	 */
	protected JButton stopButton;

	/**
	 * Button for stepping through optimization process.
	 */
	protected JButton stepButton;

	/**
	 * Box for activating step-through mode.
	 */
	protected JCheckBox stepThroughBox;

	/**
	 * Field for specifying number of iterations per step-through step.
	 */
	protected JTextField stepThroughStepSize;

	/* **********************************************************************
	 * Buttons and other components for batch mode.
	 * **********************************************************************/

	/**
	 * Configuration panel for batch mode.
	 */
	protected JPanel batchConfigPanel;
	
	/**
	 * Checkbox for enabling/disabling batch mode.
	 */
	protected JCheckBox activateBatchMode;
	
	/**
	 * Set of checkboxes to select input parameters in batch mode.
	 */
	protected Vector<JCheckBox> inputCheckBoxes;
	
	/**
	 * Components in GUI linked to the different input batch parameters.
	 */
	protected HashMap<String, ALDSwingComponent> inputProviderComps;
	
	/**
	 * Set of checkboxed to select parameters to summarize in batch mode.
	 */
	protected Vector<JCheckBox> outputCheckBoxes;
	
	/**
	 * Constructs a new control frame for an operator object.
	 * @param _op			Operator to be associated with this frame object.	
	 * @param em			Reference to the surrounding execution proxy.
	 * @throws ALDOperatorException
	 */
	public ALDOperatorControlFrame(ALDOperator _op, 
																	ALDOperatorGUIExecutionProxy em) 
			throws ALDOperatorException {
		super(_op);
		this.execProxy = em;
	}
	
	@Override
	protected void init() {
		if (this.op == null) {
			JOptionPane.showOptionDialog(
							null,
							"Problems instantiating chosen operator...\n"
									+ "Please check class implementation or contact programmer.",
							"Warning", JOptionPane.DEFAULT_OPTION,
							JOptionPane.WARNING_MESSAGE, null, this.okOption,
							this.okOption[0]);
			return;
		}

		// if no title given, generate one
		if (this.titleString == null)
			this.titleString = "ALDOperatorControlFrame: " + this.op.getName();
		
		// init the window
		this.setupWindow();
	}

	@Override
	protected HashMap<JComponent, String> setupAdditionalTabContents() {
		HashMap<JComponent, String> map = new HashMap<JComponent, String>();
		
		// setup batch configuration panel and add to map
		boolean batchModeSupport = false;
		Index<ALDAOperator, ALDOperator> indexItems = SezPozAdapter.load(
				ALDAOperator.class, ALDOperator.class);
		for (final IndexItem<ALDAOperator, ALDOperator> item : indexItems) {
			if (item.className().equals(this.op.getClass().getName())) {
				batchModeSupport = item.annotation().allowBatchMode();
				break;
			}
		}
		if (batchModeSupport) {
			this.setupBatchConfigPanel();
			// put batch mode panel into a scrollpane
			JScrollPane batchScroller = new JScrollPane(this.batchConfigPanel);
			map.put(batchScroller, "Batch Mode Configuration");
			// return filled map
			return map;
		}
		return null;
	}

	@Override
	protected Collection<JMenu> setupAdditionalMenus() {
		LinkedList<JMenu> menuList = new LinkedList<JMenu>();
		JMenu actionsM = new JMenu("Actions");
		JMenuItem itemRun = new JMenuItem("Run");
		itemRun.setActionCommand("actionsM_run");
		itemRun.addActionListener(this);
		actionsM.add(itemRun);
		return menuList;
	}

	protected void setupBatchConfigPanel() {
		this.batchConfigPanel = new JPanel();
		this.batchConfigPanel.setLayout(new BorderLayout());
		this.activateBatchMode = new JCheckBox("Activate Batch Mode");
		this.activateBatchMode.setActionCommand("batchModeToggled");
		this.activateBatchMode.addItemListener(this);

		// add list of available input parameters
		this.inputCheckBoxes = new Vector<JCheckBox>();
		this.inputProviderComps = new HashMap<String, ALDSwingComponent>();
		this.outputCheckBoxes = new Vector<JCheckBox>();
		Collection<String> inputParamNames = this.op.getInInoutNames();
		Collection<String> outputParamNames = this.op.getOutInoutNames();

		// check for which parameters you have providers
		int inputParamCount = 0;
		ALDOpParameterDescriptor descr = null;
		Collection<String> validInputParamNames = new Vector<String>();
		for (String param : inputParamNames) {
			try {
				 descr = this.op.getParameterDescriptor(param);
				if (ALDBatchInputManagerSwing.getInstance().
																				providerAvailable(descr.getMyclass())){
					++inputParamCount;
					validInputParamNames.add(param);
				}
			} catch (ALDOperatorException ex) {
				ex.printStackTrace();
      }
		}
		int outputParamCount = 0;
		Collection<String> validOutputParamNames = new Vector<String>();
		for (String param : outputParamNames) {
			try {
				 descr = this.op.getParameterDescriptor(param);
				if (ALDBatchOutputManagerSwing.getInstance().
																				providerAvailable(descr.getMyclass())){
					++outputParamCount;
					validOutputParamNames.add(param);
				}
			} catch (ALDOperatorException ex) {
				ex.printStackTrace();
      }
		}
		
		// determine size
		int paramNum = (inputParamCount > outputParamCount) ? 
																					inputParamCount : outputParamCount;

		// init header
		JPanel labelPanel = new JPanel();
		labelPanel.setLayout(new GridLayout(2+paramNum,2));
		labelPanel.add(this.activateBatchMode);
		labelPanel.add(new JLabel());
		labelPanel.add(new JLabel("Batch mode input parameters:"));
		labelPanel.add(new JLabel("Batch mode output parameters:"));
		this.batchConfigPanel.add(labelPanel,BorderLayout.NORTH);

		// temporary array for components
		JComponent [][] boxArray = new JComponent[paramNum][2]; 
		
		// add input check boxes
		ALDSwingComponent providerComp = null;
		int index = 0;
		for (String param : validInputParamNames) {
			try {
				descr = this.op.getParameterDescriptor(param);
				// check for batch iterator
				ALDBatchInputIteratorSwing bit =
						((ALDBatchInputIteratorSwing)
								(ALDBatchInputManagerSwing.getInstance().getProvider(
										descr.getMyclass(), ALDBatchInputIteratorSwing.class)));
				if (bit != null) {
					providerComp = 
						bit.createGUIElement(descr.getField(), descr.getMyclass(), 
																		null, descr);
					if (providerComp != null) {
						JPanel paramPanel = new JPanel();
						JCheckBox paramChecker = new JCheckBox(param);
						paramChecker.setSelected(false);
						paramChecker.setEnabled(false);
						paramPanel.add(paramChecker);
						paramPanel.add(providerComp.getJComponent());
						this.inputProviderComps.put(param, providerComp);
						this.inputCheckBoxes.add(paramChecker);
						// install listener
						paramChecker.setActionCommand("input_" + param);
						paramChecker.addItemListener(this);
						boxArray[index][0] = paramPanel;
					}
				}
			} catch (ALDOperatorException ex) {
				ex.printStackTrace();
			} catch (ALDBatchIOManagerException exm) {
				exm.printStackTrace();
			} catch (ALDBatchIOException exio) {
				exio.printStackTrace();
			}
			++index;
		}

		index = 0;
		for (String param : validOutputParamNames) {
			JCheckBox paramChecker = new JCheckBox(param);
			paramChecker.setSelected(false);
			paramChecker.setEnabled(false);
			this.outputCheckBoxes.add(paramChecker);
			boxArray[index][1] = paramChecker;
			++index;
		}
		
		for (index = 0; index < paramNum; ++index) {
			if (validInputParamNames.size() > index)
				labelPanel.add(boxArray[index][0]);
			else
				labelPanel.add(new JLabel(""));
			if (validOutputParamNames.size() > index)
				labelPanel.add(boxArray[index][1]);
			else
				labelPanel.add(new JLabel(""));
		}
		// add to batch mode panel
		this.batchConfigPanel.add(labelPanel,BorderLayout.CENTER);
	}
	
	/**
	 * Adds set of control buttons to the input panel.
	 */
	@Override
  protected JPanel addContextSpecificButtonPanel() {
		
		// init the panel
		JPanel runPanel = new JPanel();
		runPanel.setLayout(new GridLayout(2, 1));

		// controllable operators support step-wise execution...
		if (this.op instanceof ALDOperatorControllable) {
			this.stepThroughBox = new JCheckBox("Step-wise execution");
			this.stepThroughBox.setActionCommand("stepFlagToggled");
			this.stepThroughBox.addItemListener(this);
			JLabel label = new JLabel("                 Step size = ");
			this.stepThroughStepSize = new JTextField("1", 5);
			this.stepThroughStepSize.setEnabled(false);
			this.stepButton = new JButton("Next Step");
			this.stepButton.setActionCommand("step");
			this.stepButton.addActionListener(this);
			this.stepButton.setBounds(50, 60, 80, 30);
			this.stepButton.setEnabled(false);

			JPanel stepPanel = new JPanel();
			stepPanel.setLayout(new BoxLayout(stepPanel, BoxLayout.LINE_AXIS));
			stepPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
			stepPanel.add(this.stepThroughBox);
			stepPanel.add(label);
			stepPanel.add(this.stepThroughStepSize);
			stepPanel.add(Box.createRigidArea(new Dimension(70, 0)));
			stepPanel.add(this.stepButton);
			runPanel.add(stepPanel);
		}

		// button to re-display result frame
		this.resultDisplayButton = new JButton("Display Results");
		this.resultDisplayButton.setActionCommand("display");
		this.resultDisplayButton.addActionListener(this);
		this.resultDisplayButton.setBounds(50, 60, 80, 30);
		this.resultDisplayButton.setEnabled(false);

		// run button
		this.runButton = new JButton("Run");
		this.runButton.setActionCommand("run");
		this.runButton.addActionListener(this);
		this.runButton.setBounds(50, 60, 80, 30);

		// pause button for controllable plugins
		this.pauseButton = new JButton("Pause");
		this.pauseButton.setActionCommand("pause");
		this.pauseButton.addActionListener(this);
		this.pauseButton.setBounds(50, 60, 80, 30);
		this.pauseButton.setEnabled(false);

		// stop button for controllable plugins
		this.stopButton = new JButton("Stop");
		this.stopButton.setActionCommand("stop");
		this.stopButton.addActionListener(this);
		this.stopButton.setBounds(50, 60, 80, 30);
		this.stopButton.setEnabled(false);

		// now set up a panel to hold all buttons
		JPanel controlPanel = new JPanel();
		controlPanel
				.setLayout(new BoxLayout(controlPanel, BoxLayout.LINE_AXIS));
		controlPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
		controlPanel.add(this.resultDisplayButton);
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(Box.createHorizontalGlue());
		// controlPanel.add(validateButton);
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(this.runButton);
		// if operator is not controllable, disable control buttons
		if (this.op instanceof ALDOperatorControllable) {
			controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
			controlPanel.add(this.pauseButton);
			controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
			controlPanel.add(this.stopButton);
		}
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		runPanel.add(controlPanel);
		return runPanel;
	}

	/**
	 * Tells the surrounding proxy that a new operator has been loaded.
	 * @param newOp			New operator loaded from file.
	 */
	@Override
  protected void updateOperator(ALDOperator newOp) {
		try {
	    this.execProxy.updateOperator(newOp);
    } catch (ALDWorkflowException e) {
	    // TODO Auto-generated catch block
	    e.printStackTrace();
    }
	}
	
	/**
	 * Returns selected input parameters for batch mode.
	 * @return	List of selected inputs.
	 */
	public LinkedList<String> getBatchInputParameters() {
		LinkedList<String> inputParams = new LinkedList<String>();
		for (JCheckBox box : this.inputCheckBoxes) {
			if (box.isSelected())
				inputParams.add(box.getText());
		}
		return inputParams;
	}
	
	/**
	 * Returns an iterator for a specific input parameter in batch mode.
	 * @param param		Operator input parameter of interest.
	 * @return	Corresponding iterator.
	 * @throws ALDDataIOException
	 * @throws ALDOperatorException
	 */
	public Iterator<Object> getInputParamIterator(String param) 
			throws ALDBatchIOException, ALDOperatorException {
		return ALDBatchInputManagerSwing.getInstance().readData(null, 
				this.op.getParameterDescriptor(param).getMyclass(), 
				this.inputProviderComps.get(param));		
	}
	
	/**
	 * Returns selected output parameters for batch mode.
	 * @return	List of selected outputs to summarize.
	 */
	public LinkedList<String> getBatchOutputParameters() {
		LinkedList<String> outputParams = new LinkedList<String>();
		for (JCheckBox box : this.outputCheckBoxes) {
			if (box.isSelected())
				outputParams.add(box.getText());
		}
		return outputParams;
	}

	@Override
  public boolean quit() {
		this.stopButton.setEnabled(false);
		this.pauseButton.setEnabled(false);
		this.setVisible(false);
		return true;
	}

	/**
	 * Updates the window according to the configuration status of the operator.
	 * <p>
	 * In the window the run button color is changed according to the state of 
	 * the underlying operator node.
	 * 
	 * @param state				State of the operator and workflow node, respectively.
	 */
	public void updateNodeStatus(ALDWorkflowNodeState state) {
		// update the run button color and the tooltip
		switch(state)
		{
			case UNCONFIGURED:
				if (this.stepThroughBox== null || !this.stepThroughBox.isSelected()) {
					this.runButton.setEnabled(true);
					this.runButton.setBackground(buttonColor_unconfigured);
					this.runButton.setToolTipText("Operator not configured");
				}
				this.setStatus("Operator is unconfigured.");
				this.resultDisplayButton.setEnabled(false);
				break;
			case RUNNABLE:
				if (this.stepThroughBox== null || !this.stepThroughBox.isSelected()) {
					this.runButton.setEnabled(true);
					this.runButton.setBackground(buttonColor_runnable);
					this.runButton.setToolTipText("Operator ready to be executed");
				}
				this.setStatus("Operator is ready to run.");
				this.resultDisplayButton.setEnabled(false);
				break;
			case RUNNING:
				if (this.stepThroughBox== null || !this.stepThroughBox.isSelected()) {
					this.runButton.setBackground(buttonColor_running);
					this.runButton.setToolTipText("Operator is running...");
				}
				this.runButton.setEnabled(false);
				this.resultDisplayButton.setEnabled(false);
				this.setStatus("Operator is running...");
				// disable the parameter components
				this.operatorParameterPanel.disableComponents();
				break;
			case READY:
				if (this.stepThroughBox== null || !this.stepThroughBox.isSelected()) {
					this.runButton.setEnabled(true);
					this.runButton.setBackground(buttonColor_ready);
					this.runButton.setToolTipText(
							"Operator execution finished, results available.");
					this.resultDisplayButton.setEnabled(true);
				}
				this.setStatus("Ready.");
				// enable the parameter components
				this.operatorParameterPanel.enableComponents();
				break;
			default:
				break;
		} 
		// update the parameters
		super.updateParamConfigurationStatus(this.op.unconfiguredItems());
	}

	
	/**
	 * This method takes care of updating the GUI after reloading the operator. 
	 * @param newOp		New operator loaded from file.
	 */
	public void updateOperatorAfterConfigReload(ALDOperator newOp) {
		super.updateOperator(newOp);
	}

	/* (non-Javadoc)
	 * @see de.unihalle.informatik.Alida.gui.ALDOperatorConfigurationFrame#actionPerformed(java.awt.event.ActionEvent)
	 */
	@Override
	public void actionPerformed(ActionEvent e) {

		// get the event command
		String command = e.getActionCommand();

		// run the configured operator
		if (command.equals("run") || command.equals("actionsM_run")) {
			// check if batch mode is requested and parameters have been selected
			if (   this.activateBatchMode != null
					&& this.activateBatchMode.isSelected()) {
				try {
					this.execProxy.runWorkflowInBatchMode();
				} catch (ALDBatchIOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				} catch (ALDOperatorException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}
			else {
				if (this.op instanceof ALDOperatorControllable) {
					this.pauseButton.setEnabled(true);
					this.stopButton.setEnabled(true);
					this.repaint();
				}
				else {
					this.pauseButton.setEnabled(false);
					this.stopButton.setEnabled(false);				
				}
				// make sure that the step-wise mode is disabled
				if (this.op instanceof ALDOperatorControllable) {
					// tell the operator the requested step size via event...
					ALDConfigurationEvent confev = new ALDConfigurationEvent(this);
					confev.disableStepwiseExecution();
					this.execProxy.configureWorkflow(confev);
				}
				this.execProxy.runWorkflow();
			}
		}
		// quit, i.e. close window and kill the thread
		if (   command.equals("frame_close")
				|| command.equals("close") 
				|| command.equals("fileM_quit")) {
			this.execProxy.quit();
		}
		// stop operator execution
		else if (command.equals("stop")) {
			this.execProxy.stopWorkflow();
			this.stopButton.setEnabled(false);
			this.pauseButton.setEnabled(false);
			this.setStatus("Operator stopped.");
		}
		// pause operator execution
		else if (command.equals("pause")) {
			this.execProxy.pauseWorkflow();
			this.pauseButton.setActionCommand("continue");
			this.pauseButton.setText("Continue");
			this.stopButton.setEnabled(false);
			this.setStatus("Operator paused...");
		}
		// resume operator execution
		else if (command.equals("continue")) {
			this.execProxy.resumeWorkflow();
			this.pauseButton.setActionCommand("pause");
			this.pauseButton.setText("Pause");
			this.stopButton.setEnabled(true);
			this.setStatus("Running...");
		}
		// next step in step-wise execution
		else if (command.equals("step")) {
			if (this.execProxy.getWorkflowThreadStatus().equals(
																			WorkflowThreadStatus.THREAD_RUNNING)) {
				this.execProxy.doNextStepInWorkflow();
				this.stopButton.setEnabled(true);
			}
			// start the thread
			else {
				this.setStatus("Running...");
				try {
					String stepSizeText = this.stepThroughStepSize.getText();
					int size = Integer.valueOf(stepSizeText).intValue();
					// tell the operator the requested step size via event...
					ALDConfigurationEvent confev = new ALDConfigurationEvent(this);
					confev.setStepsize(size);
					confev.enableStepwiseExecution();
					// ... send it to operator
//					((ALDOperatorControllable) this.op)
//							.handleALDConfigurationEvent(confev);
					this.execProxy.configureWorkflow(confev);
					this.execProxy.runWorkflow();
					this.stopButton.setEnabled(true);
				} catch (Exception ex) {
					System.out.println(ex.getMessage());
					ex.printStackTrace();
				}
			}
		} 
		else if (command.equals("display")) {
			this.execProxy.showResultFrame();
		}
		else if (command.equals("helpM_about")) {
			JOptionPane
					.showOptionDialog(
							null,
							"Alida / MiToBo Graphical Operator Runner - \n"
									+ "Chosen Operator: "
									+ this.op.getName()
									+ "\n"
									+ " @2012 Martin Luther University Halle-Wittenberg",
							"About Alida / MiToBo", JOptionPane.DEFAULT_OPTION,
							JOptionPane.PLAIN_MESSAGE, null, this.okOption,
							this.okOption[0]);
		}
		else {
			// all other actions are delegated to super class
			super.actionPerformed(e);
		}
	}

	@Override
	public void handleValueChangeEvent(ALDSwingValueChangeEvent event) {
		// propegate event to super class
		super.handleValueChangeEvent(event);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
	 */
	@Override
	public void itemStateChanged(ItemEvent e) {

		String idString = null;

		// get the object affected by the event
		Object source = e.getItemSelectable();

		if (source instanceof JCheckBox) {
			JCheckBox box = (JCheckBox) source;
			idString = box.getActionCommand();
		}
		if (source instanceof JComboBox) {
			JComboBox box = (JComboBox) source;
			idString = box.getActionCommand();
		}

		// error check
		if (idString == null)
			return;
		
		// step through optimization
		if (idString.equals("stepFlagToggled")) {
			if (this.stepThroughBox.isSelected()) {
				// step-wise execution
				this.stepThroughStepSize.setEnabled(true);
				this.stepButton.setEnabled(true);
				this.runButton.setEnabled(false);
				this.runButton.setBackground(Color.GRAY);
				this.pauseButton.setEnabled(false);
			} else {
				// no step-wise execution
				this.stepThroughStepSize.setEnabled(false);
				this.stepButton.setEnabled(false);
				this.runButton.setEnabled(true);
				this.runButton.setBackground(buttonColor_runnable);
				this.pauseButton.setEnabled(false);
			}
		}
		/* ***
		 * Batch mode commands
		 * ***/
		else if (idString.equals("batchModeToggled")) {
			if (this.activateBatchMode.isSelected()) {
				for (JCheckBox box : this.inputCheckBoxes) {
					box.setEnabled(true);
				}
				for (JCheckBox box : this.outputCheckBoxes) {
					box.setEnabled(true);
				}
			}
			else {
				for (JCheckBox box : this.inputCheckBoxes) {
					// disable the box...
					box.setEnabled(false);
					// ... enable parameter component in config tab again
					String commandString = box.getActionCommand();
					String paramName = (commandString.split("_"))[1];
					this.operatorParameterPanel.setParameterNotLinked(paramName);
				}
				for (JCheckBox box : this.outputCheckBoxes) {
					box.setEnabled(false);
				}				
			}
		}
		else if (idString.startsWith("input_")) {
			// input parameter selection was changed
			String paramName = (idString.split("_"))[1];
			if (((JCheckBox)source).isSelected())
				this.operatorParameterPanel.setParameterBatchModeInput(paramName);
			else
				this.operatorParameterPanel.setParameterNotLinked(paramName);
		}
		// update GUI
		this.repaint();
	}
}
